SuperStrict
' HolzCrack by Holzchopf
' BCC #81

Framework BRL.GLMax2D
Import BRL.PNGLoader
Import BRL.Retro
Import BRL.Basic
Import BRL.Timer
Import BRL.FreeAudioAudio
Import BRL.OggLoader

' graphics output
Const GFX_WIDTH:Int		= 800
Const GFX_HEIGHT:Int		= 600
Const GFX_HERTZ:Int		= 50
' frequency analysis (bandpass filters)
Const FA_LENGTH:Int		= 1024	' not necessarily 2^x
Const FA_BANDS:Int		= 44		' also: width of big tilemap-text
Const FA_BASEFREQ:Int		= 50
Const FA_BANDWIDTH:Int		= 20
' audio stuff
Const CHN_LEFT:Int		= 0
Const CHN_RIGHT:Int		= 1
' waveform
Const WF_SAMPLEWIDTH:Int	= 2
Const WF_LENGTH:Int		= GFX_WIDTH /WF_SAMPLEWIDTH
Const WF_AMPLITUDE:Int		= 32

' set drivers and init window
apptitle = "HolzCrack"
SetGraphicsDriver(GLMax2DDriver(), GRAPHICS_BACKBUFFER | GRAPHICS_STENCILBUFFER)
SetAudioDriver("FreeAudio DirectSound")
Graphics(GFX_WIDTH, GFX_HEIGHT)

Local timer:TTimer = CreateTimer(GFX_HERTZ)

' audio
Local sample:TAudioSample = LoadAudioSample("media\BinaryBorn - bcc81.ogg")
Local music:TSound = LoadSound(sample, SOUND_LOOP)
Local channel:TChannel

' initialize freq analysis
Local freqs:Int[]
For Local f:Int = 0 Until FA_BANDS
	freqs :+ [FA_BASEFREQ +f *FA_BANDWIDTH]
Next
Local faLf:TFreqAmps = New TFreqAmps.Init(sample, CHN_LEFT, freqs, FA_LENGTH)
Local faRt:TFreqAmps = New TFreqAmps.Init(sample, CHN_RIGHT, freqs, FA_LENGTH)

' HC map
Local map:TTileMap = New TTileMap.Init("media\HOLZCHOPF.txt", faLf.amps, faRt.amps)

' waveform
Local imgwaveform:TImage = LoadImage("media\waveform.png")

' title
Local txt:String = "BCC#81*CRACKED*"
Local titletext:TWavyScrollingText = New TWavyScrollingText.Init(txt, 0, 10)
titletext.DefineMovement(0.02, 2, 1, 10)

Local ctext:String = "CODE BY HOLZCHOPF * "
ctext :+ "GRAPHICS BY HOLZCHOPF * "
ctext :+ "MUSIC BY BINARYBORN (A.K.A. HOLZCHOPF) * "
ctext :+ "BCC #81 BY BLADERUNNER * "
ctext :+ "CRACKED WITH BLITZMAX * "
ctext :+ "GRAPHICS POWERED BY OGL * "
ctext :+ "GRAPHICS CREATED WITH PHOTOSHOP AND INKSCAPE * "
ctext :+ "MUSIC CREATED WITH FL STUDIO * "
ctext :+ "SPECIAL THANKS TO KRUSBY FOR ALSO PARTICIPATING * "
ctext :+ "SUPPLEMENTARY SPECIAL THANKS TO SPIONATOM FOR ALSO PARTICIPATING * "
ctext :+ "BONUS THANKS TO FEIDER FOR ALSO PARTICIPATING * "
ctext :+ "AND EXTRA SPECIAL THANKS TO YOU FOR WATCHING AND STAYING SO LONG * "
ctext :+ "KEEP THE BLITZCODECOMPO ALIVE! * "
Local creditstext:TWavyScrollingText = New TWavyScrollingText.Init(ctext, 554, 2)
creditstext.DefineMovement(0.2, 1, 2, 10)

' background
Local imgbg:TImage = LoadImage("media\tile_bg.png")

' beat detection
Local beatv:Float
Local bgidx:Int

' start
channel = PlaySound(music)

glEnable(GL_STENCIL_TEST)

While Not (KeyDown(KEY_ESCAPE) Or AppTerminate())
	' update freq analysis
	Local pos:Int = fa_ChannelPosition(TFreeAudioChannel(channel).fa_channel)
	faLf.Update(pos)
	faRt.Update(pos)
	
	' detect beat
	Local vl:Float = faLf.peakamp
	Local vr:Float = faRt.peakamp
	beatv = Sqr(vl*vl +vr*vr) ^3
	bgidx :+ beatv *100
	
	' move scrolling texts
	titletext.Update()
	creditstext.Update()
	
	Cls
	glClear(GL_STENCIL_BUFFER_BIT)
	
	' stencil buffer setup
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE)
	glStencilFunc(GL_ALWAYS, 1, $ff)
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE)
	' stencil buffer drawing
	SetBlend MASKBLEND
	SetAlpha 1
	' draws always (every fragment) a 1 to stencil buffer
	titletext.Draw()
	' working with stencil (where stencil is 1)
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE)
	glStencilFunc(GL_EQUAL, 1, $ff)
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP)

	' draw waveform
	SetBlend ALPHABLEND
	SetOrigin 0, -62
	SetScale 1,1
	For Local c:Int = 0 To 1
		Local sdata:Float[]
		If c=0 Then
			SetColor 255,0,0
			sdata = faLf.sdata
		Else
			SetColor 0,0,255
			glColorMask(GL_FALSE, GL_FALSE, GL_TRUE, GL_TRUE)
			sdata = faRt.sdata
		EndIf
		For Local i:Int = 0 Until WF_LENGTH
			DrawImage(imgwaveform, i*WF_SAMPLEWIDTH-8, sdata[i] *WF_AMPLITUDE)
		Next
	Next
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE)

	' stop working with stencil
	' (simply invert stencil)
	glStencilFunc(GL_EQUAL, 0, $ff)
	
	' draw background
	SetBlend ALPHABLEND
	SetScale 1,1
	SetOrigin 2, -12
	For Local iy:Int = -1 To 38
		For Local ix:Int = -1 To 13
			' tile-coordinates
			' (where tx+1 is always left of tx and
			' ty+1 is always beneath ty)
			Local tx:Int = ix -iy/2 +20
			Local ty:Int = iy +tx +1
			Local phase:Int = (tx Mod 2) +(ty Mod 2)*2
			Local g:Int = 48 +Sin(bgidx *4 +phase*90) *16
			'Local g:Int = ((tx Mod 4) +(ty Mod 4)*4)*16
			SetColor g,g,g
			Local x:Int = ix *64 +(iy Mod 2) *32
			Local y:Int = iy *16
			DrawImage(imgbg, x, y)
		Next
	Next
	SetOrigin 0,0
	
	' draw big tile map text
	map.Draw()
	
	' draw credits
	SetBlend ALPHABLEND
	SetAlpha 1
	SetColor 127+Cos(bgidx)*256, 127+Cos(bgidx+120)*256, 127+Cos(bgidx+240)*256
	creditstext.Draw()
	
	WaitTimer(timer)	
	Flip 0
Wend

StopChannel(channel)

End

' frequency analysis
Type TFreqAmps
	Field sample:TAudioSample
	Field channel:Int
	Field bps:Int, cps:Int, bpc:Int
	
	Field baseperiod:Int
	Field basefreqmultiplier:Float
	Field size:Int, fsize:Float
	Field freqs:Int[]
	Field amps:Float[]
	Field peakamp:Float
	
	Field sdata:Float[]
	Field GetSampleValue:Float(pSelf:TFreqAmps, pPos:Int)
	
	' init frequency analysis
	' pSample: sample to listen
	' pChannel: channel (0=left, 1=right)
	' pFreqs: array of frequencies to listen to
	' pSize: length of filter in samples
	Method Init:TFreqAmps(pSample:TAudioSample, pChannel:Int=0, pFreqs:Int[], pSize:Int=1)
		sample = pSample
		channel = pChannel
		bps = BytesPerSample[sample.format]
		cps = ChannelsPerSample[sample.format]
		If channel >= cps Then
			channel = 0
		EndIf
		bpc = bps / cps
		' wrap GetSampleValue function
		Select sample.format
			Case SF_STEREO8, SF_MONO8
				GetSampleValue = GetSampleValue8
			Case SF_STEREO16LE, SF_MONO16LE
				GetSampleValue = GetSampleValue16LE
			Case SF_STEREO16BE, SF_MONO16BE
				GetSampleValue = GetSampleValue16BE
		End Select
		' calculate base frequency (1 Hz)
		basefreqmultiplier = 360.0 /Float(pSample.hertz)
		' copy freqs of bands
		freqs = pFreqs[..]
		size = pSize
		fsize = Float(size)
		' reserve sample data space
		sdata = New Float[size]
		' reserve amp data space
		amps = New Float[freqs.length]
		Return self
	End method
	
	' update freq analysis
	' pPos: sample position
	Method Update(pPos:Int)
		peakamp = 0
		' prepare sample data
		For Local i:Int = 0 Until size
			sdata[i] = GetSampleValue(Self, pPos+i)
		Next
		' update every band
		For Local i:Int = 0 Until freqs.length
			Local amp:Float = BandFilter(i)
			amps[i] = amp
			If amp > peakamp Then
				peakamp = amp
			EndIf
		Next
	End Method
	
	' simple band filter by multiplication of products
	' of sample data with sin and cos of band frequency
	' pBand: band (frequency index) to filter
	Method BandFilter:Float(pBand:Int)
		' (pseudo) natural weighing
		Local nw:Float = Exp(freqs[pBand]/1000.0)
		' frequency
		Local frq:Float = freqs[pBand] *basefreqmultiplier
		' trig
		Local c:Float, s:Float
		Local vcos:Float, vsin:Float
		' scan sample
		For Local i:Int = 0 Until size
			c = Cos(frq *i)
			s = Sin(frq *i)
			vsin :+ s *sdata[i]
			vcos :+ c *sdata[i]
		Next
		' amplitude
		Return Sqr(vsin*vsin +vcos*vcos) /fsize *nw
	End Method

	' returns sample value for 8 bit audio data
	' pFA: frequency analysis object
	' pPos: sample position
	Function GetSampleValue8:Float(pFA:TFreqAmps, pPos:Int)
		' condition position
		pPos :* pFA.bps
		pPos :+ pFA.channel *pFA.bpc
		pPos :Mod pFA.sample.length
		' cast sample raw data to sample value
		Local val:Int, valptr:Byte Ptr = Varptr val
		valptr[3] = pFA.sample.samples[pPos]
		val :Sar 24
		Return Float(val)/128.0
	End Function

	' returns sample value for 16 bit little endian audio data
	' pFA: frequency analysis object
	' pPos: sample position
	Function GetSampleValue16LE:Float(pFA:TFreqAmps, pPos:Int)
		' condition position
		pPos :* pFA.bps
		pPos :+ pFA.channel *pFA.bpc
		pPos :Mod pFA.sample.length
		' cast sample raw data to sample value
		Local val:Int, valptr:Byte Ptr = Varptr val
		valptr[2] = pFA.sample.samples[pPos]
		valptr[3] = pFA.sample.samples[pPos+1]
		val :Sar 16
		Return Float(val)/32768.0
	End Function

	' returns sample value for 16 bit big endian audio data
	' pFA: frequency analysis object
	' pPos: sample position
	Function GetSampleValue16BE:Float(pFA:TFreqAmps, pPos:Int)
		' condition position
		pPos :* pFA.bps
		pPos :+ pFA.channel *pFA.bpc
		pPos :Mod pFA.sample.length
		' cast sample raw data to sample value
		Local val:Int, valptr:Byte Ptr = Varptr val
		valptr[3] = pFA.sample.samples[pPos]
		valptr[2] = pFA.sample.samples[pPos+1]
		val :Sar 16
		Return Float(val)/32768.0
	End Function
End Type

' iso tile map
' with tiles' amplitude bound to freq analysis amplitudes
Type TTileMap
	Const WIDTH:Int	= 44
	Const HEIGHT:Int	= 6
	
	Field x:Int = 2, y:Int = 476
	Field tiles:TTile[]
	
	Global chnLf:Float[]
	Global chnRt:Float[]
	
	' init tile map
	' pFileName: filename of map file
	' pChnLf: array that will contain left channel freq analysis data
	' pChnRt: array that will contain right channel freq analysis data
	Method Init:TTileMap(pFileName:String, pChnLf:Float[] Var, pChnRt:Float[] Var)
		chnLf = pChnLf
		chnRt = pChnRt
		' read map
		Local mapfile:TStream = ReadFile(pFileName)
		For Local iy:Int = 0 Until HEIGHT
			Local str:String = mapfile.ReadLine()
			For Local ix:Int = 0 Until WIDTH
				' only place tile if char ain't a space
				If str[ix] <> 32 Then
					' listen to left or right channel
					Local lr:Float = iy /Float(HEIGHT)
					tiles :+ [New TTile.Init(ix,iy, lr, 1-lr)]
					' use this line for still map
					'tiles :+ [New TTile.Init(ix,iy)]
				EndIf
			Next
		Next
		mapfile.Close()
		TTile.map = Self
		Return Self
	End Method
	
	Method Draw()
		SetOrigin(x,y)
		' draw in layers
		For Local l:Int = 0 To 3
			' amplitude multiplier
			Local scale:Float = (l +1) /4.0
			SetLayerBlendings(l)		
			For Local t:TTile = EachIn tiles
				t.Draw(scale)
			Next
		Next
		SetOrigin(0,0)
	End Method

	' set blendings and colour according to layer
	' pLayer: layer to match blendings
	Function SetLayerBlendings(pLayer:Int)
		Select pLayer
		Case 0
			SetBlend ALPHABLEND
			SetColor 255,128,0
			SetAlpha 1
		Case 1
			SetBlend ALPHABLEND
			SetColor 96,0,255
			SetAlpha 0.75
		Case 2
			SetBlend LIGHTBLEND
			SetColor 128,255,240
			SetAlpha 0.75
		Case 3
			SetBlend ALPHABLEND
			SetColor 255,0,132
			SetAlpha 0.5
		End Select
	End Function
End Type

' tilemap tiles
' (with amplitudes according to frequency analysis data)
Type TTile
	Const XPERX:Int	= 16
	Const XPERY:Int	= 16
	Const YPERX:Int	= -8
	Const YPERY:Int	= 8
	
	Global img:TImage = LoadImage("media/tile.png")
	Global map:TTileMap
	
	Field ix:Int, iy:Int
	Field x:Int, y:Int
	Field listensIdx:Int
	Field listensLf:Float, listensRt:Float
	
	' init tile
	' pX: tile position in x
	' pY: tile position in y
	' pListensLf: tile sensitivity for left channel
	' pListensRt: tile sensitivity for right channel
	Method Init:TTile(pX:Int, pY:Int, pListensLf:Float=0, pListensRt:Float=0)
		ix = pX
		iy = pY
		' calculate pixel position
		x = pX *XPERX +pY *XPERY
		y = pY *YPERY +pX *YPERX
		listensLf = pListensLf
		listensRt = pListensRt
		' frequency band to listen to
		listensIdx = pX
		Return Self
	End Method
	
	' draw tile
	' pScale: amplitude multiplier
	Method Draw(pScale:Float)
		' amplitude from FA
		Local amp:Int = 200 *(map.chnLf[listensIdx] *listensLf +map.chnRt[listensIdx] *listensRt)
		' lift tile by amplitude
		DrawImage(img, x,y -amp *pScale)
	End Method
End Type

' scrolling text with time and spacial frequencies
Type TWavyScrollingText
	Const CWIDTH:Int	= 8
	
	Field txt:String
	Field y:Int
	Field scale:Int
	
	Field speed:Float
	Field foffset:Float, offset:Int
	Field dtfreq:Float, sfreq:Float, amp:Int
	Field phase:Float
	Field strlength:Int
	Field maxchars:Int
	
	' initialize text
	' pText: string
	' pY: y position of text
	' pScale: scaling factor for image
	Method Init:TWavyScrollingText(pText:String, pY:Int, pScale:Int)
		txt = pText
		y = pY
		scale = pScale
		strlength = pText.length
		' how many chars fit on screen? (add one for halfway fitting chars)
		maxchars = GFX_WIDTH /(CWIDTH *scale) +1
		Return Self
	End Method
	
	' define speed, time and spacial frequency and wave amplitude
	' pSpeed: speed in "chars" per frame
	' pTFreq: frequency in Hz
	' pSFreq: frequency in 1/GFX_WIDTH
	Method DefineMovement(pSpeed:Float, pTFreq:Float, pSFreq:Float, pAmp:Int)
		speed = pSpeed
		dtfreq = 360.0 /GFX_HERTZ *pTFreq
		sfreq = 360.0 /GFX_WIDTH *pSFreq
		amp = pAmp
	End Method
	
	' move text and update string offset
	Method Update()
		foffset :+ speed
		If foffset >= 1 Then
			foffset :- 1
			offset :+ 1
			offset :Mod strlength
		EndIf
		phase :+ dtfreq
		If phase >= 360 Then
			phase :- 360
		EndIf
	End Method
	
	' draw text
	Method Draw()
		SetScale scale, scale
		' start at offset char
		Local ic:Int = offset
		' draw only visible chars
		For Local i:Int = 0 To maxchars
			Local x:Int = (i -foffset) *CWIDTH *scale
			Local y:Int = y +Sin(x *sfreq -phase) *amp
			DrawText(Chr(txt[ic]), x, y)
			ic :+ 1
			ic :Mod strlength
		Next
		SetScale 1,1
	End Method
End Type


